#*****************************************************************************************
#                                  LED-Matrix-Server
#                                  =================
#                                  Uwe Berger; 2009
#
# Es wird eine Matrix mit den Dimensionen der Variablen gvar(dx) und gvar(dy) gezeichnet.
# Via TCP (z.B. telnet localhost 12345) koennen Kommandos gesendet werden. Folgende
# Kommandos sind implementiert (Beschreibung/Parameter sind dem Quelltext zu entnehmen):
# 
# help
# get_xy
# clear_all
# set_point <x> <y> <color>
# set_sign  <x> <y> <sign> <color>
# set_line  <x1> <y1> <x2> <y2> <color>
# set_led   <LED-Array-String>
#
# Es sollten im Script alle Vorkehrungen getroffen worden sein, um die Remote-
# kommunikation sicher zu gestalten und nur obige Kommandos sollten via Netzwerk
# ausfuehrbar sein...
#
# ---------
# Have fun!
#
#*****************************************************************************************

source sign_lib.tcl;                            # Font-Bibliothek

#**** einige Variable ****
set gvar(dx) 48;                                # x-Dimension				
set gvar(dy) 32;                                # y-Dimension
set gvar(raster) 14;                            # Abstand der Punkte untereinander
set gvar(led_radius) [expr $gvar(raster) / 2];  # Durchmesser ;-) der LEDs
set gvar(led) {};                               # LED-Feld (wird spaeter initialisiert)
set gvar(server_port) 12345;                    # Serverport
set gvar(max_colors) 4;                         # Anzahl Farbzustaende
set gvar(led_colors) {gray5 red4 red3 red};     # Farbwerte 0...3

#**** Oberflaeche generieren ****
proc gui_init {} {
	global gvar
	wm title . "LED-Matrix ($gvar(dx)x$gvar(dy))" 
	frame .f2
	canvas	.f2.matrix -width [expr ($gvar(dx) + 1) * $gvar(raster) + $gvar(raster) / 2] \
			-height [expr ($gvar(dy) + 1) * $gvar(raster) + $gvar(raster) / 2] \
			-bg black -bd 0
	pack .f2 -fill both
	grid .f2.matrix -row 0 -column 0 -columnspan 3
}

#**** LED-Feld initialisieren ****
proc led_init {} {
	global gvar
	for {set i 0} {$i < [expr $gvar(dx) * $gvar(dy)]} {incr i} {
		set gvar(led) [linsert $gvar(led) end 0] 
	}
}

proc draw_matrix {} {
	global gvar
	.f2.matrix delete all
	for {set y 0} {$y < $gvar(dy)} {incr y} {
		for {set x 0} {$x < $gvar(dx)} {incr x} {
			set i [expr $x + $y * $gvar(dx)]
			set color [lindex $gvar(led_colors) [lindex $gvar(led) $i]]
			.f2.matrix create oval [expr $gvar(raster)  * ($x + 1)] \
						[expr $gvar(raster)  * ($y + 1)] \
						[expr ($gvar(raster) * ($x + 1)) + $gvar(led_radius)] \
						[expr ($gvar(raster) * ($y + 1)) + $gvar(led_radius)] \
						-outline $color -fill $color
		}
	}
}

#**** eine einzelne LED setzen ****
proc draw_point {x y val} {
	global gvar
	if {($x < $gvar(dx)) && ($y < $gvar(dy)) && ($val < $gvar(max_colors))} {
		set i [expr $x + $y * $gvar(dx)]
		set gvar(led) [lreplace $gvar(led) $i $i $val]
	}
}

#**** ein (vordefiniertes) Zeichen schreiben ****
proc draw_sign {letter x y c} {
	global gvar sign fdx fdy
	for {set i 0} {$i < [llength $sign($letter)]} {incr i} {
		if {[lindex $sign($letter) $i]} {
			set py [expr int($i/$fdx) + $y]
			set px [expr ($i % $fdx) + $x]
			draw_point $px $py $c
		}
	}
}

#**** eine Linie zeichnen ****
proc draw_line {x1 y1 x2 y2 c} {
	set dx  [expr abs($x2 - $x1)]
	set dy  [expr abs($y2 - $y1)]
	set my1 [expr -$y1]
	set my2 [expr -$y2]
	set st  [expr {$dy > $dx}]

	if {$st} {
		set d $x1;  set x1 $my1;   set my1 $d
		set d $x2;  set x2 $my2;   set my2 $d
		set d $dx;  set dx $dy;    set dy $d
	}

	set i1 [expr 2 * $dy]
	set d  [expr $i1 - $dx]
	set i2 [expr $d - $dx]
	set yi 1
	if {$x1 > $x2} {
		set x $x2
		set y $my2
		set xe $x1
	if {$my2 > $my1} {set yi -1}
	} else {
		set x $x1
		set y $my1
		set xe $x2
		if {$my1 > $my2} {set yi -1}
	}
	while {$x <= $xe} {
		if {$st} {
			set xp $y
			set yp [expr -$x]
		} else {
			set xp $x
			set yp [expr -$y]
		}
		draw_point $xp $yp $c
		incr x
		if {$d > 0} {
			incr y $yi
			incr d $i2
		} else {
			incr d $i1
		}
	}
}

#************************ Server-Kommunikations-Routinen ***************************
#----- accept -----
proc accept {sock addr port} {
	fconfigure $sock -buffering line
	fileevent  $sock readable [list receive $sock $addr $port]
	puts "-> Accept connection: $addr/$port\n"

}                                   

#----- receive -----
proc receive {sock addr port} {
	global safe
	if {[eof $sock] || [catch {gets $sock line}]} {
		puts "-> close connection: $addr/$port\n"
		close $sock
	} else {
		puts "-> receive from connection ($addr/$port/$sock): >$line<"
		set cmd [lindex $line 0]
		puts "-> command receive: $cmd"
		if {$line != ""} {
			if {[catch {$safe eval $line $sock}]} {
				puts "-> unknown command!"
			}
		}
	}
}

#***** Serverkommandos *******
#----- help -------
proc help {sock} {
	global gvar
	puts $sock "-----Hilfe-----"
	puts $sock "help - diese Hilfeseite"
	puts $sock "get_xy - gibt Dimension der Matrix zurueck"
	puts $sock "clear_all - alle LEDs loeschen"
	puts $sock "set_point <x> <y> <color> - ein LED setzen"
	puts $sock "set_sign  <x> <y> <sign> <color> - ein Zeichen ausgeben"
	puts $sock "set_line  <x1> <y1> <x2> <y2> <color> - eine Linie setzen"
	puts $sock "set_led   <LEDs> - alle LEDs als eine Zeichenfolge uebergeben 0111030..."
	puts $sock "<color> - 0...3; 0=aus; 3=ganz hell"
	flush $sock
}

#----- get_xy -----
proc get_xy {sock} {
	global gvar
	puts $sock "set_xy $gvar(dx) $gvar(dy)"
	flush $sock
}

#----- set_led -----
proc set_led {led sock} {
	global gvar
	set gvar(led) [split $led {}]
	draw_matrix
	update
	puts $sock "update_ok"
	flush $sock
}

#----- set_point -----
proc set_point {x y val sock} {
	global gvar
	draw_point $x $y $val
	draw_matrix
	update
	puts $sock "update_ok"
	flush $sock
}

#----- set_sign -----
proc set_sign {x y sign c sock} {
	global gvar
	draw_sign $sign $x $y $c
	draw_matrix
	update
	puts $sock "update_ok"
	flush $sock
}

#----- set_line -----
proc set_line {x1 y1 x2 y2 c sock} {
	global gvar
	draw_line $x1 $y1 $x2 $y2 $c
	draw_matrix
	update
	puts $sock "update_ok"
	flush $sock
}

#----- clear_all -----
proc clear_all {sock} {
	global gvar
	set gvar(led) {}
	led_init
	draw_matrix
	update
	puts $sock "update_ok"
	flush $sock
}


#********************************************************
#***************** Main *********************************
#********************************************************
gui_init
led_init
draw_matrix

# sicheren Interpreter erzeugen (immerhin gehen wir uebers Netz...)
set safe [interp create -safe]
$safe alias get_xy    get_xy
$safe alias set_point set_point
$safe alias set_sign  set_sign
$safe alias set_line  set_line
$safe alias clear_all clear_all
$safe alias set_led   set_led
$safe alias help      help

# TCP/IP-Kommunikationskanal oeffnen und dort lauschen
set gvar(sock) [socket -server accept $gvar(server_port)]

